-------------------------------------------------------------------------
-- Joseph Zambreno
-- Department of Electrical and Computer Engineering
-- Iowa State University
-------------------------------------------------------------------------


-- BubbleSort.vhd
-------------------------------------------------------------------------
-- DESCRIPTION: This file contains an implementation of BubbleSort that
-- sorts 8 inputs using individual comparator units and assuming a
-- serialized bus.
--
--
-- NOTES:
-- 8/24/06 by JAZ::Design created.
-------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity BubbleSort is

  generic(gBUS_WIDTH    : natural := 8;
          gDATA_WIDTH   : natural := 8);
  port(iCLK             : in std_logic;
       iLOAD            : in std_logic;
       iDATA            : in std_logic_vector(gBUS_WIDTH-1 downto 0);
       oDATA            : out std_logic_vector(gBUS_WIDTH-1 downto 0);
       oDONE            : out std_logic);

end BubbleSort;

architecture dataflow of BubbleSort is
  component comparator
    generic(gDATA_WIDTH : natural := 8);
    port(iCLK           : in std_logic;
         iA             : in std_logic_vector(gDATA_WIDTH-1 downto 0);
         iB             : in std_logic_vector(gDATA_WIDTH-1 downto 0);
         oL             : out std_logic_vector(gDATA_WIDTH-1 downto 0);
         oH             : out std_logic_vector(gDATA_WIDTH-1 downto 0));
  end component;

  type DATAarray is array (1 to 8) of std_logic_vector(gDATA_WIDTH-1 downto 0);
  signal sARRAY1, sARRAY2, sARRAY3, sARRAY4 : DATAarray;
  signal sLOAD1, sLOAD2, sLOAD3, sLOAD4 : std_logic;
  
begin

  
  ---------------------------------------------------------------------------
  -- Level 1: Each input is compared to its nearest neighbor
  ---------------------------------------------------------------------------
  g_Comp1: for i in 1 to 4 generate
    g_Comp1_i: Comparator
      generic map(gDATA_WIDTH   => gDATA_WIDTH)
      port MAP(iCLK             => iCLK,
               iA               => sARRAY1(2*i-1),
               iB               => sARRAY1(2*i),
               oL               => sARRAY2(2*i-1),
               oH               => sARRAY2(2*i));
  end generate;
    
  ---------------------------------------------------------------------------
  -- Level 2: Each higher value is compared to neighbors highest value
  ---------------------------------------------------------------------------
  g_Comp2a: for i in 1 to 2 generate
    g_Comp2a_i: Comparator
      generic map(gDATA_WIDTH   => gDATA_WIDTH)
      port MAP(iCLK             => iCLK,
               iA               => sARRAY2(i),
               iB               => sARRAY2(i+2),
               oL               => sARRAY3(2*i-1),
               oH               => sARRAY3(2*i));
  end generate;
    
  g_Comp2b: for i in 3 to 4 generate
    g_Comp2b_i: Comparator
      generic map(gDATA_WIDTH   => gDATA_WIDTH)
      port MAP(iCLK             => iCLK,
               iA               => sARRAY2(i+2),
               iB               => sARRAY2(i+4),
               oL               => sARRAY3(2*i-1),
               oH               => sARRAY3(2*i));    
  end generate;
    
  ---------------------------------------------------------------------------
  -- Level 3: Each higher value is to the other half
  ---------------------------------------------------------------------------
  g_Comp3: for i in 1 to 4 generate
    g_Comp3_i: Comparator
      generic map(gDATA_WIDTH   => gDATA_WIDTH)
      port MAP(iCLK             => iCLK,
               iA               => sARRAY3(i),
               iB               => sARRAY3(i+4),
               oL               => sARRAY4(2*i-1),
               oH               => sARRAY4(2*i));    
  end generate;

  ---------------------------------------------------------------------------
  -- Process that collects the input on iLOAD and starts the pipeline.
  ---------------------------------------------------------------------------
  process(iCLK, iLOAD, iDATA)
    variable i : integer := 1;
    variable j : integer := 1;
    variable vLOAD : std_logic := '0';
  begin
    if rising_edge(iCLK) then
      vLOAD := '0';
      if (iLOAD = '1') then
        sARRAY1(j)(gBUS_WIDTH*i-1 downto gBUS_WIDTH*(i-1)) <= iDATA;
        i := i + 1;
        if (i > gDATA_WIDTH/gBUS_WIDTH) then
          i := 1;
          j := j + 1;
          if (j > 8) then
            vLOAD := '1';
            j := 1;
          end if;
        end if;
      end  if;
    end if;
    sLOAD1 <= vLOAD;
  end process;
  
  ---------------------------------------------------------------------------
  -- Process that collects the sorted output and serializes it on the bus
  ---------------------------------------------------------------------------
  process(iCLK, sLOAD4, sARRAY4)
    variable i : integer := 1;
    variable j : integer := 1;
    variable vDONE : std_logic := '0';
  begin
    if rising_edge(iCLK) then
      if (sLOAD4 = '1') then
        vDONE := '1';
      end if;
      if (vDONE = '1') then
        oDONE <= '1';
        oDATA <= sARRAY4(j)(gBUS_WIDTH*i-1 downto gBUS_WIDTH*(i-1));
        i := i + 1;
        if (i > gDATA_WIDTH/gBUS_WIDTH) then
          i := 1;
          j := j + 1;
          if (j > 8) then
            j := 1;
            vDONE := '0';
          end if;
        end if;
      else
        oDONE <= '0';
      end if;
    end if;
    --oDONE <= vDONE;
  end process;
    
  ---------------------------------------------------------------------------
  -- Process that pipelines the sLOADx signals 
  ---------------------------------------------------------------------------

  process(iCLK, sLOAD1, sLOAD2, sLOAD3)
  begin
    if rising_edge(iCLK) then
      sLOAD4 <= sLOAD3;
      sLOAD3 <= sLOAD2;
      sLOAD2 <= sLOAD1;
    end if;
  end process;
  
end dataflow;
